#
# Copyright 2016 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

class Api::V4::ReportsController < Api::V3::ReportsController
  # Preprocess the query results.
  #
  # @param [DecoratedPluckyQuery] query_results Query results
  # @param [String] version API version string.
  # @return [Hash] Query results suitable for rendering as json.
  def processed_results(query_results, version)
    transforms = {String => [:interpolate_ctf_urls],
                  ActivitySummary => [:translate_summary_to_decorated_hash]}
    walk_and_transform(super, transforms)
  end

  # Aggregate source display data from all the ActivitySummaries found in the
  # query results.
  #
  # @param [DecoratedPluckyQuery, Array, Hash] results Data structure to search
  # @return [Hash] Hash map of display data, keyed by source id.
  def add_metadata_to_results(results, sources={}, work_item_urls={})
    iterator = results.respond_to?(:all) ? results.all : results
    iterator.each do |item|

      if item.is_a? ActivitySummary
        # Construct and return URLs for work item summaries
        work_item_urls["#{item.id}"] = item.url if item.is_a? WorkItemSummary

        next if sources.key? item.source.id
        # Build basic source display information:
        sources[item.source.id] = {
            _type: item.source._type,
            _links: {
                type_icon: {
                    href: icon_url(item.source),
                    type: icon_type(item.source)
                },
            },
            name: item.source.display_name,
            background_color: icon_color(item.source),
            source_association_key: item.source.source_association_key
        }

        # Add the schema name, if applicable:
        sources[item.source_id][:schema_display_name] = item.source.schema.name if item.is_a? CustomSummary

        # Inject information on the source icon, if present (i.e., Tracker sources):
        source_icon = source_icon_url(item.source)
        if source_icon
          sources[item.source.id][:_links][:source_icon] = {
              href: source_icon,
              type: source_icon_type(item.source)
          }
        end
      else
        add_metadata_to_results(item, sources, work_item_urls) if [Array, Hash].include? item.class
      end
    end
    { source_display_data: sources, work_item_urls: work_item_urls }
  end

  # Walk a data structure, performing any specified proccessing on nodes.
  #
  # @param [DecoratedPluckyQuery, Hash, Array, String, Object] results Node to walk.
  # @param [Hash<Class: [Method,...]>] transforms Hash of methods to run, keyed by class.
  # @return [Hash, Array, String, Object] Node procesed by the transformations.
  def walk_and_transform(results, transforms={})
    # Force execution of query
    results = results.respond_to?(:all) ? results.all : results

    # Perform all supplied transformations:
    transforms.each do |cls, methods|
      methods.each {|mthd| results = self.method(mthd).call results } if results.is_a? cls
    end

    # Walk the structure (destructively):
    if results.is_a? Array
      results.map! { |item| walk_and_transform(item, transforms) }
    elsif results.is_a? Hash
      results.each { |k, v| results[k] = walk_and_transform(v, transforms) }
    end
    results
  end

  # Perform CTF %VAR% interpolations on the provided string.
  #
  # @param [String] text The string to interpolate
  # @return [String] Transformed string.
  def interpolate_ctf_urls(text)
    CtfInterpolations.interpolate_ctf_tokens(text)
  end

  # Translates an ActivitySummary to Hash and decorates with additional information.
  #
  # @param [ActivitySummary] summary_obj An activity summary object.
  # @return [Hash] Decorated and converted hash representation of input.
  def translate_summary_to_decorated_hash(summary_obj)
    return_hash = {'source_association_key' => summary_obj.source.source_association_key}
    if summary_obj.is_a? CustomSummary
      return_hash['schema_display_name'] = summary_obj.source.schema.name
    end
    summary_obj.attributes.merge!( return_hash )
  end

end
