#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# Represents the users
class User
  include MongoMapper::Document

  DEVISE_PARAMS = :registerable, :database_authenticatable, :recoverable, :rememberable, :trackable,
      :lockable, :omniauthable

  # We want to skip validation in teamforge-based usage, because email can be non-unique, and passwords are unnecessary.
  DEVISE_PARAMS << :validatable unless Rails.configuration.teamforge[:enabled]

  devise *DEVISE_PARAMS

  attr_accessible :uid, :provider, :name, :email, :password, :password_confirmation, :authentication_token

  # Omniauthable
  key :uid, String, required: true
  key :provider, String, required: true
  key :name, String, required: true
  validates_uniqueness_of :uid, :scope => :provider

  # Token Authenticable
  key :authentication_token

  ## Database authenticatable
  key :email, String, :required => true
  key :encrypted_password, String, :required => true

  ## Recoverable
  key :reset_password_token, String
  key :reset_password_sent_at, Time

  ## Rememberable
  key :remember_created_at, Time

  ## Trackable
  key :sign_in_count, Integer, :default => 0
  key :current_sign_in_at, Time
  key :last_sign_in_at, Time
  key :current_sign_in_ip, String
  key :last_sign_in_ip, String

  ## Encryptable
  # key :password_salt, String

  ## Confirmable
  #key :confirmation_token,   String
  #key :confirmed_at,         Time
  #key :confirmation_sent_at, Time
  #key :unconfirmed_email,    String # Only if using reconfirmable

  ## Lockable
  key :failed_attempts, Integer # Only if lock strategy is :failed_attempts
  key :unlock_token,    String  # Only if unlock strategy is :email or :both
  key :locked_at,       Time

  # Token authenticatable
  # key :authentication_token, String

  ## Invitable
  # key :invitation_token, String

  key :created_at, Time
  key :updated_at, Time

  #Store the user's specified timezone, default to Pacific zone until set at users first login
  key :time_zone, String, default: 'America/Los_Angeles'

  has_many :user_permissions
  key :refreshed_permissions_at, default: Time.at(0)

  key :super_user, Boolean, default: false

  validates_associated :user_permissions

  embedded_callbacks_off

  # Retrieve a list of active projects that the user has the provided permissions on.
  #
  # @param [Symbol] permission The permission to check against.
  # @return [Array[Project]] The list of projects that are active and the user has the provided permission in.
  def projects_with_permission(permission)
    if @super_user
      Project.active_projects
    else
      Project.active_projects_by_id(project_ids_with_permission(permission))
    end
  end

  # Retrieve a list of project_ids for projects the user has the provided permissions on.
  #
  # @param [Symbol] permission The permission to check against.
  # @return [Array[BSON id]] The list of project ids for projects the user has the provided permission in.
  def project_ids_with_permission(permission)
    if @super_user
      Project.active_projects.map(&:id)
    else
      user_permissions.select do |perm|
        Ability.permissions_implying(permission).include? perm.name.to_sym
      end.collect(&:project_id).uniq
    end
  end

  # Find out if a permission exists for the user on a project.
  #
  # @param [Symbol] permission The permission to check.
  # @params [BSON id] project_id The project to check.
  # @return [Boolean] Asserts if the user have the permission on the specified project.
  def has_permission_for_project_id?(permission, project_id)
    project_ids_with_permission(permission).include? project_id
  end

  def has_write_permission_for_project_id?(project_id)
    has_permission_for_project_id?(:pipeline_create, project_id) || has_permission_for_project_id?(:pipeline_edit, project_id)
  end

  # Queries the CTF for the projects that the user is a member of and returns an array of their ids
  # @param [TeamForge::Service] ctf_service a teamforge CTF service
  # @return [Array] the ctf projects ids
  def ctf_projects_ids(ctf_service)
    ctf_projects = ctf_service.get_project_list_for_user(user_id: uid, fetch_hierarchy_path: false, include_group_membership: false)[:multi_ref]
    ctf_projects.to_a.map{|project| project[:id]}
  end

  # Returns the ctf project_ids for the projects that user is a member of and which also have ORC as an IAF app
  # @param [TeamForge::Service] ctf_service a teamforge CTF service
  # @return [Array]
  def iaf_projects_ids(ctf_service)
    ctf_projects_ids(ctf_service) & Project.active_projects_ctf_ids
  end

  # Pings CTF for IAF permissions and returns an array of permissions
  # @param [String] ctf_rbac, CTF session id
  # @param [String] ctf_project_id, CTF project id
  # @return [Array] array of permission names if the user has any permissions in this project
  # @return [nil] if the  user does not have any permissions in this project
  def iaf_permissions_for_project(ctf_rbac, ctf_project_id)
    link_id = TeamForge::ProjectConfiguration.find_by_ctf_project_id(ctf_project_id).get_link_plug_id(ctf_rbac.session_id)
    permissions = ctf_rbac.get_integrated_app_permissions_for_project(project_id: ctf_project_id, link_id: link_id)[:get_integrated_app_permissions_for_project_return]
    if permissions
      permissions_array = ([] << permissions).flatten #need this because CTF returns a string for a single permission or array for multiple
      permissions_array.map{|perm| perm.downcase.gsub(/ +/,"_")}
    else
      []
    end
  end

  # Construct from first principles the user's list of permissions to CTF projects.
  # Gets the list of project ids for this user from CTF, and then for each of them
  # asks for the user's permissions to that project.  The resultant permissions set
  # becomes the #user_permissions attribute.
  # It is implemented as a simple map/reduce operation, mapping from
  # projects:Array[permissions]:Array[UserPermissions] and reducing that to a single array.
  #
  # @param [TeamForge::Session] ctf_session A CTF SOAP session with teamforge and rbac SOAP endpoints.
  # @return [Array[UserPermission]] A list of UserPermission objects for each permission/project pair.
  def ctf_permissions(ctf_session)
    iaf_projects_ids(ctf_session.teamforge).reduce([]) do |accum, ctf_project_id|
      project_config = TeamForge::ProjectConfiguration.find_by_ctf_project_id(ctf_project_id)

      accum + iaf_permissions_for_project(ctf_session.rbac, ctf_project_id).map do |perm|
        UserPermission.new(name: perm, project_id: project_config.project_id)
      end
    end
  end

  # Obtains API permissions for this user
  # @return [Array] of UserPermission
  def api_permissions
    teamforge_config = Rails.configuration.teamforge
    session = CtfService.create_teamforge_session.tap { |tf| tf.login(teamforge_config[:iaf][:admin_user], teamforge_config[:iaf][:admin_pass]) }
    plug_id = session.pluggable.get_plug_id_by_base_url(base_url: Rails.configuration.teamforge[:iaf][:orc_url])

    permissions = []
    iaf_projects_ids(session.teamforge).each do |ctf_project_id|
      logger.debug "Obtaining permissions for CTF project: #{ctf_project_id}"
      project_config = TeamForge::ProjectConfiguration.find_by_ctf_project_id(ctf_project_id)
      api_perms = %w(reporting_api)
      api_perms.each do |api_permission|
        if ctf_permission_for_project?(api_permission, plug_id, project_config, session)
          permissions << UserPermission.new(name: api_permission, project_id: project_config.project_id)
        end
      end
    end
    permissions
  rescue => e
    logger.warn "Unable to get IAF plug id from base_url '#{Rails.configuration.teamforge[:iaf][:orc_url]}'; the Orchestrate IAF may not be set up correctly.  Contact support."
    logger.warn e.message
    []
  end

  # Test whether this user has a given permission on a project
  # @param [String] permission to test
  # @param [ProjectConfiguration] project_config to test
  # @param [TeamForge::Session] ctf_session to use
  # @return [Boolean] true or false, with errors logged
  def ctf_permission_for_project?(permission, plug_id, project_config, ctf_session)
    logger.debug "Testing permission '#{permission}' on #{project_config.try(:link_plug_id)}"
    operation = ["PluggableLink", plug_id, permission.humanize, permission.humanize].join('.')

    #noinspection RubySimplifyBooleanInspection
    # this client method returns boolean true or a soap response hash, which we want to interpret as false.
    ctf_session.teamforge.has_general_permission(username: uid,
                                                 project_id: project_config.ctf_project_id,
                                                 operation: operation,
                                                 object_id: project_config.link_plug_id) == true
  rescue => e
    logger.warn "Unable to test permission '#{permission}' on (#{project_config.try(:ctf_project_id)} / #{project_config.try(:link_plug_id)})"
    logger.warn e.message
    false
  end

  # Called by #refresh_permissions; this updates the user's permissions with permissions retrieved
  # from CTF. When no CTF SOAP session is provided, the permissions are cleared and only API-access permissions
  # are populated
  # @param [TeamForge::Session] ctf_session A CTF SOAP session for current user with teamforge and rbac SOAP endpoints.
  def teamforge_refresh_permissions(ctf_session = nil)
    update_attribute :user_permissions, ctf_session ? ctf_permissions(ctf_session) : api_permissions
  end

  # Refreshes permissions and sets the next update time to 20 hours from the current time
  # @param [TeamForge::Session || Hash] black_box, right now is either a CTFSession object (for CTF authorization) or a hash (for local auth)
  def refresh_permissions(black_box = nil)
    permission_method = "#{provider}_refresh_permissions"
    send(permission_method, black_box) if respond_to? permission_method
    update_attribute :refreshed_permissions_at, Time.now.utc
  end

  # Finds a user with the given omniauth authentication information or creates them if one does not already exist.
  # 'auth' is an AuthHash provided by OmniAuth; provider is the current OmniAuth
  # strategy in use.
  def self.find_or_create_user(auth)
    provider = auth.provider
    uid = auth.uid
    user = User.first(provider: provider, uid: uid)
    user ||= User.create(email: auth.info.email, name: auth.info.name,
                         uid: uid, provider: provider, password: Devise.friendly_token[0, 20])
    user.tap { |u| u.ensure_authentication_token }
  end

  #TODO migrate to the recommended devise implementation (https://gist.github.com/josevalim/fb706b1e933ef01e4fb6)
  # devise no longer supports this type of authentication so adding this to allows us to move forward
  def ensure_authentication_token
    if authentication_token.blank?
      self.authentication_token = generate_authentication_token
    end
  end

  # Updates the user name and email, since it may have changed in teamforge.
  def after_login(auth)
    self.name = auth.info.name
    self.email = auth.info.email
    save!
  end

  # Load a set of permissions from the provided file (in YAML format) and apply them to all the users; if a user isn't
  # specifically mentioned in the file, use the 'default' user permissions.
  # @param [String] permissions_file The path to the YAML file containing permissions
  def self.apply_permissions(permissions_file)
    stored_permissions = YAML.load_file(permissions_file)

    # Now we grant permissions to users according to the user_auth_store.yml file.
    User.all.each do |user|
      auth = stored_permissions[user.uid] || stored_permissions['default']
      user.refresh_permissions auth
    end
  end

  # Given a set of permissions, clear the user's permissions, and add all provided project permissions.
  #
  # @param [Hash] auth A collection of permissions in the form of { 'project_name' => 'permission'}
  def localauth_refresh_permissions(auth = nil)
    return unless auth
    self.user_permissions = []
    auth['project_permissions'].each do |project_name, permissions|
      project = Project.find_by_name(project_name)
      permissions.each do |permit|
        self.user_permissions << UserPermission.new(name: permit, project: project)
      end if project
    end
    save
  end

  private
  def generate_authentication_token
    loop do
      token = Devise.friendly_token
      break token unless User.where(authentication_token: token).first
    end
  end
end
