#!/opt/collabnet/ruby/bin/ruby
#
# Copyright 2015 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

# install.rb
# The TeamForge Orchestrate installer
#
require 'getoptlong'
require 'rubygems'
require 'capistrano'
require 'capistrano/cli'
require 'fileutils'
require File.expand_path('../../lib/installer/helpers.rb',  __FILE__)
require File.expand_path('../../lib/installer/questions.rb',  __FILE__)
require File.expand_path('../../lib/installer/config.rb',  __FILE__)
require File.expand_path('../../lib/installer/physician.rb',  __FILE__)

begin

  options = {}
  getopts = GetoptLong.new(
      ['-r', '--repo', GetoptLong::REQUIRED_ARGUMENT],
      ['-l', '--local', GetoptLong::NO_ARGUMENT],
      ['-f', '--file', GetoptLong::REQUIRED_ARGUMENT]
  )

  getopts.each do |opt, arg|
    case opt
      when '-f'
        options[:file] = arg
      when '-r'
        options[:repo] = arg
      when '-l'
        options[:local] = ORCHESTRATE_LOCAL_OPTIONS_FILE
    end
  end

  unless is_root
    puts "ERROR: You must be root to run the #{ORCHESTRATE_NAME} installer!"
    exit(10)
  end

  puts "CollabNet #{ORCHESTRATE_NAME} Installer\n\n"
  puts "INFO: CollabNet #{ORCHESTRATE_NAME} Installer\n\n"

  puts "INFO: Ensuring local ssh keys exist and are authorized"
  ensure_localhost_keys

  # We used to die here if found, but we can now attempt incremental installation, so just issue a warning here.  This
  # test really indicates that "we got far enough in the install that defaults.yml was generated"
  #
  # TODO: replace this test with checking the state marker for the orchestrate task?
  if existing_install
    puts "WARNING: An existing install of #{ORCHESTRATE_NAME} has been detected; attempting to continue"
  end

  # Make sure we have a safe place to put all our stuff:
  FileUtils.mkdir_p(INSTALLER_BASE_DIR, mode: 0600)

  Asky.send(:include, InstallerQuestions)
  questions = Asky.new(SiteOptions.new([options[:local], options[:file]]).options)

  # Ask them all the things
  [:mongodb, :rabbitmq, :nginx, :orchestrate, :teamforge].each { |subsection| questions.send(subsection) }

  # Check for any existing (e.g., rabbitmq) settings that should be merged into the answers
  # TODO: should we just do this before asking the questions, and if we find good data, just
  # flag it as an existing system?  The install is skipped below, if an existing_system is
  # present.
  questions.answers.deep_merge!(get_existing_settings(questions.answers))
  questions.answers.merge!({'rpm_repo' => options[:repo]})

  # We are done manipulating answers; save out the file here.
  open(INSTALLER_LAST_OPTIONS_FILE, 'w'){ |file|
    file.write(SiteOptions.to_options_string(questions.answers, sanitize=false))
  }

  # Create the yaml file that the rest of the installation will use.
  installer_yaml = create_installer_yaml(questions.answers)

  # Save installer config out to YAML for backend
  open(INSTALLER_YAML_CONFIG_FILE, 'w') { |file| file.write(installer_yaml) }
  File.chmod(0600, INSTALLER_YAML_CONFIG_FILE)  # Keep it secret, keep it safe

  begin
    puts "\nINFO: Beginning #{ORCHESTRATE_NAME} installation and configuration ..."
    reinstall = questions.answers['orchestrate']['operation'] == 'clean_reinstall'

    cap_tasks = []

    unless reinstall
      cap_tasks << ['orchestrate', ['orchestrate:orchestrate:check_prefix_availability', '-s', 'role=orchestrate']]
    end

    VALID_ROLES.each do |role|
      unless questions.answers[role.to_s].has_key?('existing_system')
        operation = (questions.answers[role] && questions.answers[role][:operation]) || 'install'
        cap_tasks << [role.to_s, ["orchestrate:#{role}:#{operation}", '-s', "role=#{role}"]]
      end
    end

    current_phase = nil

    # Do all the stuff for real.
    cap_tasks.each do |role, cap_params|
      current_phase = cap_params[0]
      if questions.answers[role.to_s].has_key?('existing_system')
        puts "INFO: Role #{role} is using an existing system; skipping installation"
        next
      end
      state = check_marker(cap_params)
      if state == :completed
        puts "INFO: Detected previously completed phase #{current_phase} (role=#{role}); skipping"
        next
      end
      if state == :started
        puts "WARNING: Previously started but uncompleted phase #{current_phase} detected (role=#{role}); continuing"
        # TODO: keep going?  Load the existing answers?  Or present the diff here?
      end
      set_marker(cap_params, answers: questions.answers, state: :started)
      puts "INFO: Starting phase: #{current_phase} (role=#{role})"
      # TODO: are we reliably catching failure here?
      Capistrano::CLI.parse(cap_params).execute!
      set_marker(cap_params, state: :completed)
      puts "INFO: Completed phase: #{current_phase} (role=#{role})"
    end
  rescue Exception => e # TODO: move this exception to the end?
    puts "\nERROR: An exception occurred during installation. The exception was:"
    puts e.to_s
    puts "\nTRACE:"
    puts e.backtrace

    doctor = Physician.new(e, current_phase)

    if doctor.has_diagnosis
      puts doctor.format_diagnosis
    end

    puts "\nThe last options supplied have been saved to #{INSTALLER_LAST_OPTIONS_FILE}"
    puts "You may edit and re-supply those answers by re-running this script and passing the argument '-f #{INSTALLER_LAST_OPTIONS_FILE}'."
    exit(14)
  end

  # Write out the equivalent of the site-options.conf file used to generate the current installation
  # (without passwords):
  open(ORCHESTRATE_SITE_OPTIONS_FILE, 'w'){ |file| file.write(SiteOptions.to_options_string(questions.answers))}
  File.chmod(0600, ORCHESTRATE_SITE_OPTIONS_FILE)

  # TODO: do we really want to unlink/clear anything here?  It stands as a good record of what was run; if need be,
  # we can remove the entire contents of INSTALLER_BASE_DIR at this point.
  #
  #File.delete(INSTALLER_YAML_CONFIG_FILE)
  #clear_markers

  puts "\nInstallation of #{ORCHESTRATE_NAME} complete."
  puts "\nA copy of the settings used to configure this installation have been saved to #{ORCHESTRATE_SITE_OPTIONS_FILE}."
  puts "\nThe #{ORCHESTRATE_NAME} application server is up and running.  This phase of #{ORCHESTRATE_NAME} set up"
  puts "is complete. Please review the post-install documentation for any additional instructions on completing the"
  puts "installation and configuration with TeamForge."
  puts "\n"
rescue Interrupt
  diaf(12)
end
