#!/bin/sh
#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

#
# Orchestrate installer wrapper
#
set -e

REPO_URL='http://orchestrate.collab.net/RHEL/$releasever/x86_64'
RUBY_ROOT="/opt/collabnet/ruby"
ORCH_ROOT="/opt/collabnet/orchestrate"
ORCH_LICENSE="$ORCH_ROOT/LICENSE.txt"
ORCH_LICENSE_ACCEPTED="/opt/collabnet/orchestrate/tmp/installer/orc_license_accepted"

# Prints basic usage information
usage()
{
cat << EOF
usage: $0 [options]

This script starts the TeamForge Orchestrate installer.

OPTIONS:
   -h,--help                        Show this message
   -r,--repo <repo-url>             Specify an alternate repository URL for Orchestrate
   -f,--file <config>               Specify a orchestrate site-options answer file
   -l,--local                       Assume all components should be installed on localhost
   --install-version <version>      Install a specific version or revision of the orchestrate-core RPM, e.g. '1.0.0' or '1.0.1-64'
EOF
}

# Installs the Collabnet software Repo
install_repo()
{
  REPO_FILE='/etc/yum.repos.d/CollabNet-Orchestrate.repo'
  if [ -f "$REPO_FILE" ]; then
    return
  fi

  echo "Installing the CollabNet-Orchestrate yum repository..."

cat << EOF > $REPO_FILE
[collabnet]
name=Collabnet TeamForge Orchestrate Package Repository
enabled=1
baseurl=$REPO_URL
gpgcheck=0
EOF
}

# Rolls back the install, generally called after a problem occurs
rollback()
{
  # Removing libyaml actually removes everything related to Orchestrate
  yum -y remove collabnet-orchestrate-core collabnet-libyaml collabnet-redis
  rm /etc/yum.repos.d/CollabNet-Orchestrate.repo
}

# Check for the existence of an answers file, and report no file found and exit if it's provided but not there
nofile()
{
  if [ ! -z "$1" -a ! -r "$1" ]; then
    echo "The answers file you specified ($1) does not exist.  Either provide the proper path to the answers file or run the installer in interactive mode." >&2
    exit 1
  fi
}

# Parse the parameters for options
TEMP_OPTS=`getopt -o hr:f:l -l help,skip-license,repo:,file:,install-version:,local -n "$0" -- "$@"`

# Exit if parsing failed
if [ $? != 0 ] ; then
  exit 1
fi

#Set $@ to the reordered parameters from getopt 
eval set -- "$TEMP_OPTS"

# Parse any options
while true 
  do
    case "$1" in
      -h|--help)
          usage
          exit
          ;;
      -r|--repo)
          REPO_URL=$2
          REPO_OPTION="$1 $REPO_URL"
          shift 2
          ;;
      -f|--file)
          ANSWERS_FILE=$2
          shift 2
          ;;
      -l|--local)
          LOCAL_OPTION='--local'
          shift
          ;;
      --install-version)
         INSTALL_VERSION=$2
         shift 2
         ;;
      --skip-license)
         SKIP_LICENSE=true
         shift
         ;;
      --)
          shift
          break
          ;;
      *)
          echo "Parse error! Aborting"
          exit 1
          ;;
    esac
  done

if [ $(id -u) -ne 0 ]; then
  echo 'You must be root to run the Orchestrate installer!' >&2
  exit 1
fi

# Add collabnet repo to the system
install_repo

# Install the Orchestrate RPM
echo "Downloading and extracting Orchestrate..."
PACKAGE=collabnet-orchestrate-core
if [ -n "$INSTALL_VERSION" ]; then
  PACKAGE=$PACKAGE-$INSTALL_VERSION
fi
yum -y install $PACKAGE

# Ensure any linking issues are resolved (artf250601)
ldconfig /opt/collabnet/yaml/lib/


# Find the full path for the answers file
if [ ! -z "$ANSWERS_FILE" ]; then
  export ANSWERS_FILE
  CANONICALIZED_ANSWERS=`$RUBY_ROOT/bin/ruby -e "print File.expand_path(ENV['ANSWERS_FILE'])"`
  nofile $CANONICALIZED_ANSWERS
  ANSWERS_OPTION="-f $CANONICALIZED_ANSWERS"
fi

# User must agree to the license to continue
if [ ! -f "$ORCH_LICENSE" ]; then
  echo "Orchestrate license agreement text is missing, the installer can't run without this."
  exit 2
fi

not_valid=1

if [ -f "$ORCH_LICENSE_ACCEPTED" ]; then
  echo "INFO: Previous acceptance of the license agreement found; assumes that you agree to the license agreement located here: $ORCH_LICENSE"
  sleep 2
  not_valid=0
elif [ -z "$SKIP_LICENSE" ]; then
  echo
  less -e -F -m "$ORCH_LICENSE"
  echo
else
  echo "INFO: Use of the --skip-license flag assumes that you agree to the license agreement located here: $ORCH_LICENSE"
  sleep 2
  not_valid=0
fi

while [ $not_valid -eq 1 ]; do
  read -p "Do you agree to the terms of the license?: "

  if [[ $REPLY =~ ^[Yy] ]]; then
    not_valid=0
    mkdir -p `dirname $ORCH_LICENSE_ACCEPTED` && chmod 0600 `dirname $ORCH_LICENSE_ACCEPTED`
    touch $ORCH_LICENSE_ACCEPTED
  elif [[ $REPLY =~ ^[Nn] ]]; then
    echo "Orchestrate cannot be installed without acceptance of the license agreement."
    echo "Cleaning up..."
    sleep 5 
    rollback
    exit 5
  else
    echo -e "Please answer y or n...\n"
  fi
done

echo -e "\nStarting the installer...\n"

cd $ORCH_ROOT
env PATH=$RUBY_ROOT/bin:/sbin:$PATH bundle exec $ORCH_ROOT/bin/install.rb $ANSWERS_OPTION $REPO_OPTION $LOCAL_OPTION
