#!/bin/bash
#
# Upgrade script for Orchestrate
#

SERVICE_NAMES=(orchestrate mongodb rabbitmq)

ORCHESTRATE_INIT='/etc/init.d/orchestrate'
ORCHESTRATE_PACKAGE='collabnet-orchestrate-core'
ORCHESTRATE_ROOT='/opt/collabnet/orchestrate'

RABBITMQ_INIT='/etc/init.d/collabnet-rabbitmq-server'
RABBITMQ_PACKAGE='collabnet-rabbitmq'

MONGODB_INIT='/etc/init.d/collabnet-mongod'
MONGODB_PACKAGE='collabnet-mongodb'

DEFAULTS_YAML=${ORCHESTRATE_ROOT}/config/defaults.yml

RUBY_ROOT=/opt/collabnet/ruby

# Ensure that PATH will have chkconfig, etc.
export PATH=${PATH}:/sbin:/usr/sbin:/usr/local/sbin

# Usage statement
usage() {
  cat << EOF | sed -e 's/^ //g'
  usage: $0 orchestrate|mongodb|rabbitmq|all
  or
  usage: $0 --help

  The script will upgrade the specified application.

  OPTIONS:
     -h,--help      Show this message
EOF
}

upgrade_orchestrate() {
  set -e
  echo -e "\nShutting down existing Orchestrate instance..."
  # Work around a conflict betweeen old-fashined init.d scripts and
  # systemd-managed services by skipping redirect while shutting down:
  export SYSTEMCTL_SKIP_REDIRECT=1
  ${ORCHESTRATE_INIT} stop
  unset SYSTEMCTL_SKIP_REDIRECT

  echo -e "\nUpgrading  Orchestrate ..."
  #these rpm yum commands work around an upgrade bug in libyaml and ruby installed for Orchestrate versions 1.0.*
  set +e
  rpm -e --nodeps --nopostun collabnet-libyaml-0.1.4
  rpm -e --nodeps --nopostun collabnet-ruby-1.9.3_p392
  set -e 
  yum clean all
  yum -y install "collabnet-libyaml"
  yum -y install "collabnet-ruby"
  # Ensure any linking issues are resolved (artf250601)
  ldconfig /opt/collabnet/yaml/lib/
  yum -y update "collabnet-orchestrate-core"

  pushd ${ORCHESTRATE_ROOT}

  echo -e "\nCompiling Orchestrate assets... (this may take a while)"
  su -c "bundle exec rake assets:clean" - orchestrate
  su -c "RAILS_RELATIVE_URL_ROOT=/orc RAILS_ENV=production bundle exec rake assets:precompile" - orchestrate
  echo -e "\nRunning database migrations"
  su -c "bundle exec rake db:migrate" - orchestrate
  # copy init scripts:
  cp config/templates/init.d/orchestrate/* /etc/init.d/
  cp config/templates/init.d/nginx/collabnet-nginx.erb /etc/init.d/collabnet-nginx
  # copy any systemd service definitions:
  [ -d /etc/systemd/system ] && cp config/templates/systemd/*service /etc/systemd/system 
  # TODO: There may at some point be the need to modify/upgrade/refresh nginx's
  # conf file, but artf240114 wants the preservation of existing port, etc.
  # configuration, so for the time being do nothing here.
  declare -a services=("collabnet-nginx" "collabnet-redis" "orchestrate" "orchestrate-ctf-jobs" "orchestrate-queue-commit" "orchestrate-queue-build" "orchestrate-queue-review" "orchestrate-queue-work-item" "orchestrate-queue-custom")
  for i in "${services[@]}"
  do
    chkconfig --del "$i"
    chkconfig --add "$i"
  done
  popd

  # Call restart (rather than start) to ensure that services have been completely
  # stopped under their current service definitions, as those definitions may have
  # changed during the course of the upgrade: 
  echo -e "\nUpgrade complete, restarting Orchestrate ..."
  ${ORCHESTRATE_INIT} restart
  set +e

  echo -e "\nUpdating IAF configuration ..."
  su -c "bundle exec rake orchestrate:update_iaf_configuration" - orchestrate
  echo -e "\nIAF configuration updated ..."
  set +e
}

upgrade_rabbitmq() {
  set -e
  echo -e "\nShutting down rabbitmq service ..."
  ${RABBITMQ_INIT} stop

  echo -e "\nUpgrading rabbitmq ..."
  yum clean all
  yum -y update ${RABBITMQ_PACKAGE}

  echo -e "\nUpgrade complete, starting rabbitmq service ..."
  ${RABBITMQ_INIT} start
  set +e
}

upgrade_mongodb() {
  set -e
  echo -e "\nShutting down mongodb service ..."
  ${MONGODB_INIT} stop

  echo -e "\nUpgrading mongodb ..."
  yum clean all
  yum -y update ${MONGODB_PACKAGE}

  echo -e "\nUpgrade complete, starting mongodb service ..."
  ${MONGODB_INIT} start
  set +e
}

check_prerequisites() {
  check_required_commands
  check_defaults_yml    
}

check_required_commands() {
  echo -e "\nTesting required commands in environment ..."

  for cmd in rpm yum chkconfig; do 
    which $cmd > /dev/null;
    if [ $? -ne 0 ]; then
      echo "FATAL: command $cmd not found in PATH.  Please install $cmd or ensure that it is in PATH."
    exit 1
    fi
  done
}

check_defaults_yml() {
  echo -e "\nChecking defaults.yml configuration ..."

  if [ ! -f "$DEFAULTS_YAML" ]; then
    echo "FATAL: no $DEFAULTS_YAML file found." 
    exit 1
  fi

  # Does it look like Rabbit is advertising localhost?:
  egrep '^\s*uri:\s*("|'"'"')?amqps?:.*(127\.0\.0\.1|localhost)' $DEFAULTS_YAML >/dev/null 2>&1 

  if [ $? -eq 0 ]; then
    now=`date +%Y%m%d%H%M%S`
    cp -p $DEFAULTS_YAML ${DEFAULTS_YAML}.${now}
    update_rabbit_host
  fi
}

# Interactively edit the contents of the defaults.yml file in-place.
update_rabbit_host() {
  ruby_interpreter=`([[ -x ${RUBY_ROOT}/bin/ruby ]] && echo ${RUBY_ROOT}/bin/ruby)||echo '/opt/collabnet/ruby/bin/ruby'`
  script_name='/tmp/update_rabbit_host'  
  # sed to strip the indentation here
  cat << EOF | sed -e 's/^  //g' > $script_name
  #!$ruby_interpreter

  require 'yaml'
  require 'uri'

  STDOUT.sync = true 
  STDIN.sync = true 

  defaults_file = ARGV[0] || '/opt/collabnet/orchestrate/config/defaults.yml'

  unless File.file? defaults_file
      puts "FATAL: no such file #{defaults_file}"
      exit 1
  end

  data = YAML.load(File.read(defaults_file))
  uri = URI(data['rabbitmq']['production']['uri'])
  loopback = ['127.0.0.1', 'localhost']

  if loopback.include?(uri.host)
      puts "ERROR: the value for the RabbitMQ host is a loopback interface."
      while loopback.include?(uri.host)
          print ">> Please supply the FQDN of the rabbit host: "
          host = STDIN.gets.strip
          uri.host = host unless host.empty?
      end
      data['rabbitmq']['production']['uri'] = uri.to_s
      File.open(defaults_file, 'w') { |f|
          f.write data.to_yaml
      }
      puts "#{defaults_file} successfully updated."
  end
EOF
  chmod 755 $script_name
  $script_name $DEFAULTS_YAML || exit 1
}

### Live code below:

# Complain and exit if not root
if [ $(id -u) -ne 0 ]; then
  echo 'You must be root to run the Orchestrate upgrader!' >&2
  exit 1
fi

# Show help message if -h or --help is passed
case "$1" in
  -h|--help|"")
      usage
      exit 1
      ;;
  *)
      SERVICE_NAME=$1
      ;;
esac

# Copy STDOUT and STDIN to the upgrade.log
TIMESTAMP=$(date +"%Y%m%d%H%M%S")
LOG_FILE="`dirname $0`/upgrade_${TIMESTAMP}.log"
exec > >(tee $LOG_FILE)
exec 2>&1

# make sure that the correct service name is passed as an argument
[[ $SERVICE_NAME =~ ^(orchestrate|mongodb|rabbitmq|all)$ ]]
if [ $? -ne 0 ]; then
  usage
  exit 1
fi

check_prerequisites

# Upgrade the service
case $SERVICE_NAME in
  "all")
    upgrade_rabbitmq
    upgrade_mongodb
    upgrade_orchestrate
    ;;
  "orchestrate")
    upgrade_orchestrate
    ;;
  "rabbitmq")
    upgrade_rabbitmq
    ;;
  "mongodb")
    upgrade_mongodb
    ;;
esac
