#
# Copyright 2015 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

require_relative 'config'

# Class to diagnose common installer exceptions.
class Physician

  # TODO: extract the hash as external constants if the physician's library ever grows to any reasonable extent.
  SYMPTOMS = {
    # Prefix phase:
    /\(ns1:LoginFault\)\s*Session is invalid or timed out/i => {
        cause: "Your #{TEAMFORGE_APP_NAME} license is invalid or expired.",
        actions: ["Please confirm that your #{TEAMFORGE_APP_NAME} license is current."]
    },

    /\(ns1:LoginFault\)\s*Error logging in.\s*Please verify the username and password/i => {
        cause: "The credentials supplied for administrative access to #{TEAMFORGE_APP_NAME} are invalid.",
        actions: ["Edit #{INSTALLER_LAST_OPTIONS_FILE} file with corrected credentials, and re-run the installer with" +
                  " that file supplied (-f #{INSTALLER_LAST_OPTIONS_FILE})"]
    },

    /Connection refused - connect/i => {
        cause: "It appears that #{TEAMFORGE_APP_NAME} is not running.",
        actions: ["Please ensure that #{TEAMFORGE_APP_NAME} is running and accessible."]
    },

    /HTTP error \(503\)/i => {
        cause: "It appears that #{TEAMFORGE_APP_NAME}'s JBoss application server is not running.",
        actions: ["Please ensure that #{TEAMFORGE_APP_NAME} is running and accessible."]
    },

    /Some prefixes.*already exists in Teamforge/i => {
        cause: "#{TEAMFORGE_APP_NAME} already contains the prefixes the installation was trying to create." +
            "  This can happen when the installer is run repeatedly, when #{TEAMFORGE_APP_NAME} contains an IAF entry from a previous" +
            ' Orchestrate installation, or when another non-Orchestrate IAF is using the same prefix.',
        actions: ["A site admin should remove the existing IAF entry from #{TEAMFORGE_APP_NAME}, and resume this installation.",
                  'Either the existing IAF should change its prefix(es), or Orchestrate should use custom prefix(es).']
    },

    # RabbitMQ phase:
    /failed:.*collabnet-rabbitmq-server start/i => {
        cause: 'RabbitMQ experienced some permissions problems when attempting to start up the server.',
        actions: ['Check directory and logfile permissions and ownership; confirm that paths mention in the stack' +
                  ' trace exist and have the correct ownership and permissions.']
    },

    /failed:.*rabbitmqctl add_user/i => {
        # FIXME: this shouldn't occur any more, but if it did, what should the user do?
        cause: 'A problem occurred when attempting to add users to RabbitMQ.' +
            '  This problem can occur when attempting to add a user that already exists,' +
            ' or when passwords are out-of-sync.',
        actions: ["Confirm that the RabbitMQ password used by Orchestrate (in #{ORCHESTRATE_DEFAULTS_YAML_FILE})" +
                  " is the same as the one in the RabbitMQ application (#{RABBITMQ_PASSWORDS_FILE})"],
    },

    # MongoDB phase:
    /Error checking admin account/i => {
        cause: 'The installer was unable to confirm that a user could connect to the Mongo database.' +
            '  This problem can occur when attempting to add a user that already exists,' +
            ' or when passwords are out-of-sync.',
        actions: ["Confirm that the Mongodb password used by Orchestrate (in #{ORCHESTRATE_DEFAULTS_YAML_FILE})" +
                  " is the same as the one in the Mongodb application (#{MONGODB_PASSWORDS_FILE})"],
    },

    /mongo.*db.getSiblingDB.*addUser/i => {
        cause: 'The installer was unable to add a programmatic user to the Mongo database.' +
            '  This problem can occur when attempting to add a user that already exists,' +
            ' or when passwords are out-of-sync.',
        actions: ["Confirm that the Mongodb password used by Orchestrate (in #{ORCHESTRATE_DEFAULTS_YAML_FILE})" +
                  " is the same as the one in the Mongodb application (#{MONGODB_PASSWORDS_FILE})"],
    },

    # Nginx phase:
    /collabnet-nginx.*(start|stop)/i => {
        cause: 'Nginx webserver could not be started.',
        actions: ['Test and confirm that the generated configuration file is valid (e.g.,' +
                  " `sudo #{NGINX_HOME_DIR}/sbin/nginx -c #{NGINX_CONFIG_FILE} -t`)"],
    },

    # Orchestrate phase:
    /failed:.*orchestrate start/i => {
        cause: 'Orchestrate application could not be started.' +
            '  This could be due to a dirty reinstallation of Orchestrate, resulting in uname/uid mismatch,' +
            ' or other permissions errors.',
        actions: ['Check user entries, and permissions in the rabbitmq application directory.']
    },

    /failed:.*CtfConfigurer.new.create_iaf_application/i => {
        cause: "Orchestrate could not create the necessary IAF entry in #{TEAMFORGE_APP_NAME}." +
            "  This can be due to: proxy misconfiguration between #{TEAMFORGE_APP_NAME} and Orchestrate, certificate validation" +
            ' errors, and possibly a pre-existing IAF with the same base.',
        actions: ["Check #{TEAMFORGE_APP_NAME}'s proxy configuration",
                  "If Orchestrate is using SSL, confirm that the certificates can be validated by #{TEAMFORGE_APP_NAME}"],
    },

    # General failures:
    /failed:.*chown root/i => {
        cause: 'SELINUX is being enforced.',
        actions: ['Disable SELINUX for the Orchestrate host.'],
    },

    /connection failed for:.*Net::SSH::AuthenticationFailed/i => {
        cause: 'SSH credentials supplied were not suitable to access the deployment host',
        actions: ['Check the user/password/SSH key being used to deploy to the host'],
    },

    /failed:.*yum -y install collabnet-/i => {
        cause: 'The yum repository needed to complete the installation does not exist or is unreachable',
        actions: ['Check the yum repo definition (usually in /etc/yum.repos.d/), that it is reachable, and active'],
    },
  }

  # TODO: does phase have any value?  Remove?
  def initialize(exception, phase=nil)
    @exception = exception
    @phase = phase
    @diagnosis = []

    # Create a diagnosis based upon the exception
    ex_str = @exception.to_s
    SYMPTOMS.keys.each do |key|
      @diagnosis << SYMPTOMS[key] if ex_str =~ key
    end
  end

  # Tell me, doctor - is it bad?
  #
  # @return [Boolean] Boolean indicating if a diagnosis was arrived at.
  def has_diagnosis
    !@diagnosis.empty?
  end

  # Tell me, doctor - how bad is it?
  #
  # @return [Array[Hash]] Array of two-key hashes, containing the cause and actions.
  def diagnosis
    @diagnosis.empty? ? nil : @diagnosis
  end

  # Alright, doctor - give it to me straight.
  #
  # @return [String] String containing diagnosis.
  def format_diagnosis
    str = "\nDIAGNOSIS:\n"
    diagnosis.each do |condition|
      str << " POSSIBLE CAUSE:\n"
      str << " * #{condition[:cause]}\n"
      str << "\n SUGGESTED ACTIONS:\n"
      condition[:actions].each { |action| str << " * #{action}\n" }
    end
    str << "\n"
    str
  end
end