#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

require 'securerandom'
require 'teamforge'
require File.expand_path(File.join(File.dirname(__FILE__), '..', '..', 'ctf','ctf_default_prefixes'))

# Converts any type of hash to a generic Ruby hash and stringifies the keys
# @param [Object] hash_to_convert The hash to be converted
# @return [Hash]
def to_ruby_hash(hash_to_convert)
  hash_to_convert.inject({}) { |result, (key, value)|
    value = to_ruby_hash(value) if value.is_a?(Hash)
    result[(key.to_s rescue key) || key] = value
    result
  }
end

module Installer::Plugins::Orchestrate

  JOBS = %w(orchestrate orchestrate-ctf-jobs orchestrate-queue-commit orchestrate-queue-build orchestrate-queue-work-item orchestrate-queue-custom orchestrate-queue-review)

  LOGROTATE_TEMPLATE_FILE = File.join('config', 'templates', 'logrotate', 'orchestrate')

  # Main entry point to get all of orchestrate installed and configured properly
  # @param [Symbol] install_type - A symbol of the installation type, either :install, or :clean_reinstall
  def setup(install_type = :install)
    validate(install_type)
    install_initd
    setup_logrotate
    generate_random_csrf_token unless install_type == :clean_reinstall
    setup_defaults_yml_file(fetch(:defaults_yml))
    precompile_assets
    create_indexes
    run_migrations
    create_site_wide_linked_app(fetch(:defaults_yml)) unless install_type == :clean_reinstall
    setup_defaults_yml_file(fetch(:defaults_yml))
    start
    create_iaf_app unless install_type == :clean_reinstall
    check_daemons
  end

  # perform validation on required field for orchestrate module
  def validate(install_type = :install)
    validate_defaults_yml
    validate_prefixes_conflict(fetch(:defaults_yml)) unless install_type == :clean_reinstall
  end

  # Validation for the defaults_yml file for orchestrate  
  def validate_defaults_yml
    #TODO figure this one out
    field = :defaults_yml
    role = current_role

    field_value = installer.fetch_installer_config(role, field)

    set(field, field_value)
  rescue IndexError
    raise "Missing #{role} #{field}"
  end

  # validate if prefixes already exists in CTF. If not an exact match we abort
  # @param [Hash] defaults_yml - the defaults_yml hash
  def validate_prefixes_conflict(defaults_yml)

    begin
      ctf_admin_user = defaults_yml['teamforge']['production']['iaf']['admin_user']
      ctf_admin_password = defaults_yml['teamforge']['production']['iaf']['admin_pass']
      ctf_server = defaults_yml['teamforge']['production']['server']
    rescue NoMethodError
      raise 'Misconfigured default yaml for the teamforge section'
    end

    sess = new_teamforge_session(ctf_server)
    sess.login(ctf_admin_user, ctf_admin_password)
    response = sess.pluggable.get_integrated_app_prefixes

    override_prefix = defaults_yml['teamforge']['production']['iaf']['prefix']
    orch_prefixes = override_prefix.nil? ? CtfDefaultPrefixes::PREFIXES.values : [override_prefix]

    #SOAP returns nil for empty set, string for a single entry, and array with multiple entries
    #Here we turn it into empty array, array with single entry, or array with multiple entries
    teamforge_prefixes = [response[:get_integrated_app_prefixes_return]].compact.flatten

    conflicts = teamforge_prefixes & orch_prefixes

    raise "Some prefixes (#{conflicts.join(', ')}) already exists in Teamforge. Cannot install Orchestrate.." if !conflicts.empty?
  end

  # create a new teamforcge_session
  def new_teamforge_session(ctf_server)
    TeamForge::Session.new( ctf_root_url: ctf_server,
                            session_id: nil,
                            wsdl_directory: File.join('config', 'wsdl'))
  end

  # return the current role
  # @return [Symbol] the current role
  def current_role
    installer.current_role
  end

  # Copy init.d scripts to /etc/init.d
  def install_initd
    path_prefix = File.join(ORCHESTRATE_HOME_DIR, 'config', 'templates', 'init.d', 'orchestrate')

    JOBS.each do |job|
      installer.install_initd(File.join(path_prefix, job), job)
    end
  end

  # install the logrotate for orchestrate
  def setup_logrotate
    installer.install_logrotate(LOGROTATE_TEMPLATE_FILE, 'orchestrate')
  end

  # write the default yaml generated by the front end to the config file
  # needed by our rails app
  def setup_defaults_yml_file(defaults_yml)
    tmp_default_filename = '/tmp/default.yml'
    File.open(tmp_default_filename, 'w+'){|f| f << to_ruby_hash(defaults_yml).to_yaml}

    if installer.remote_file_exists? ORCHESTRATE_DEFAULTS_YAML_FILE
      run "#{sudo} mv #{ORCHESTRATE_DEFAULTS_YAML_FILE} #{ORCHESTRATE_DEFAULTS_YAML_FILE}.$(date +#{BACKUP_TIMESTAMP_FORMAT})"
    end

    run "#{sudo} mv #{tmp_default_filename} #{ORCHESTRATE_DEFAULTS_YAML_FILE}"
    run "#{sudo} chown root:#{ORCHESTRATE_USER} #{ORCHESTRATE_DEFAULTS_YAML_FILE}"
    run "#{sudo} chmod 0640 #{ORCHESTRATE_DEFAULTS_YAML_FILE}"
  end

  # Start orchestrate and background jobs
  def start
    run "#{sudo} #{SYSTEM_INITD_DIR}/orchestrate start"
  end

  # Stop orchestrate and background jobs
  def stop
    run "#{sudo} #{SYSTEM_INITD_DIR}/orchestrate stop"
  end

  # precompile the assets
  def precompile_assets
    puts "\n\nThe next step may take a few minutes, depending on the capabilities of your system.  Please be patient...\n\n"
    run "#{sudo} runuser -s /bin/bash orchestrate -c \"cd /opt/collabnet/orchestrate && RAILS_RELATIVE_URL_ROOT=/orc RAILS_ENV=production bin/bundle exec /opt/collabnet/ruby/bin/rake assets:precompile\""
  end

  # create database indexes
  def create_indexes
    run "#{sudo} runuser -s /bin/bash orchestrate -c \"cd /opt/collabnet/orchestrate && RAILS_RELATIVE_URL_ROOT=/orc RAILS_ENV=production bin/bundle exec /opt/collabnet/ruby/bin/rake db:index\""
  end

  # Runs migrations to ensure any migration generated data is added to the database.
  def run_migrations
    run "#{sudo} runuser -s /bin/bash orchestrate -c \"cd /opt/collabnet/orchestrate && RAILS_RELATIVE_URL_ROOT=/orc RAILS_ENV=production bin/bundle exec /opt/collabnet/ruby/bin/rake db:migrate\""
  end

  # create the site wide linked app in CTF
  # @param [Hash] defaults_yml - the defaults_yml hash
  def create_site_wide_linked_app(defaults_yml)
    app_id = @config.capture "#{sudo} runuser -l orchestrate -c \"cd #{ORCHESTRATE_HOME_DIR}; RAILS_ENV=production bin/bundle exec rails runner 'puts CtfConfigurer.new.create_site_wide_linked_application'\""

    defaults_yml['teamforge']['production']['linked_app_id'] = app_id.strip
    set(:defaults_yml, defaults_yml)
  end

  # creatae the iaf app
  def create_iaf_app
    run "#{sudo} runuser -l orchestrate -c \"cd #{ORCHESTRATE_HOME_DIR}; RAILS_ENV=production bin/bundle exec rails runner 'puts CtfConfigurer.new.create_iaf_application'\""
  end

  # Replaces the contents of 'secret_token.rb' with a new token
  # @param [String] file_path Path to the secret_token.rb file
  # @param [Integer] num_bytes Length of the token in bytes
  def generate_random_csrf_token(file_path="#{ORCHESTRATE_HOME_DIR}/config/initializers/secret_token.rb", num_bytes=128)
    token = SecureRandom.hex(num_bytes)
    file_contents = "Orchestrate::Application.config.secret_token = '#{token}'\n"

    open(ORCHESTRATE_TMP_SECRET_TOKEN, 'w') { |file| file.write(file_contents) }
    run "#{sudo} mv #{ORCHESTRATE_TMP_SECRET_TOKEN} #{file_path}"
    run "#{sudo} chown root:root #{file_path}"
  end

  # check if the daemons are running
  # Raise if a daemon is found to not be running
  def check_daemons
    (JOBS - ['orchestrate']).each do |job|
      out = @config.capture "#{sudo} #{File.join(SYSTEM_INITD_DIR, job)} status"
      raise "#{job} daemon is not started/not running" if out =~ /is stopped/
    end   
  end
end

Capistrano.plugin :app_server, Installer::Plugins::Orchestrate
