#
# Copyright 2015 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

module Installer::Plugins::Rabbitmq

  RABBITMQ_INIT_SCRIPT_NAME = 'collabnet-rabbitmq-server'
  RABBITMQ_INIT_SCRIPT = File.join('config', 'templates', 'init.d', 'rabbitmq', RABBITMQ_INIT_SCRIPT_NAME)

  RABBITMQ_LOGROTATE_NAME  = 'collabnet-rabbitmq-server'
  RABBITMQ_LOGROTATE_TEMPLATE = File.join('config', 'templates', 'logrotate', 'rabbitmq-server')


  # entry point to setting up in correct order rabbitmq
  def setup
    validate
    install_initd
    setup_logrotate
    enable_plugins
    start
    add_user(fetch(:orc_user), fetch(:orc_password))
    set_permission_for_orc_user(fetch(:orc_user))
    reset_guest_password(fetch(:guest_password))
    status
    verify_connection(fetch(:ssh_hosts), fetch(:guest_password))
  end

  # Validate the minimum requirement to run
  def validate
    validate_orchestrate_user
    validate_orchestrate_password
    validate_orchestrate_guest_password
    validate_host
  end

  # return the current role
  # @return [Symbol] the current role
  def current_role
    installer.current_role
  end

 # Validation for the rabbitmq orchestrate user
  def validate_orchestrate_user
    validate_field_presence(:orc_user)
  end

 # Validation for the rabbitmq orchestrate password
  def validate_orchestrate_password
    validate_field_presence(:orc_password)
  end

  # Validation for the rabbitmq guest password
  def validate_orchestrate_guest_password
    validate_field_presence(:guest_password)
  end

  def validate_host
    validate_field_presence(:host)
    installer.validate_host_not_localhost()
  end

  # wrapper to isolate dependency from base and repeated code for validation of field present
  # @param [Symbol] field - the field to check presence of
  def validate_field_presence(field)
    installer.validate_field_presence(current_role, field)
  end

  # install an /etc/init.d/rabbitmq-server file.
  def install_initd
    installer.install_initd(RABBITMQ_INIT_SCRIPT, RABBITMQ_INIT_SCRIPT_NAME)
  end

  # install the logrotate config file to /etc/logrotate.d/rabbitmq-server
  def setup_logrotate
    installer.install_logrotate(RABBITMQ_LOGROTATE_TEMPLATE, RABBITMQ_LOGROTATE_NAME)
  end

  # enables rabbitmq plugins
  def enable_plugins
    run "#{sudo} #{RABBITMQ_HOME_DIR}/sbin/rabbitmq-plugins enable rabbitmq_management"
    run "#{sudo} chmod 644 #{RABBITMQ_HOME_DIR}/etc/rabbitmq/enabled_plugins"
  end

  #starts rabbitmq
  def start
    run "#{sudo} #{SYSTEM_INITD_DIR}/#{RABBITMQ_INIT_SCRIPT_NAME} start"
  end

  #stop rabbitmq
  def stop
    run "#{sudo} #{SYSTEM_INITD_DIR}/#{RABBITMQ_INIT_SCRIPT_NAME} stop"
  end

  # get the rabbitmq status 
  def status
    run "#{sudo} #{SYSTEM_INITD_DIR}/#{RABBITMQ_INIT_SCRIPT_NAME} status"
  end

  # add a new user to rabbitmq
  #
  # @param [String] user - the user to create
  # @param [String] password - the password for that user
  def add_user(user, password)
    unless user_exists?(user)
      run "cd #{RABBITMQ_HOME_DIR} && #{sudo} runuser -l #{RABBITMQ_USER} -c '#{RABBITMQ_HOME_DIR}/sbin/rabbitmqctl add_user #{user} #{password}'"
      write_user_password(user, password)
    else
      logger.info "User already added"
    end
  end

  # set permission for orc user
  # @param [String] user - the user to set permissions on
  def set_permission_for_orc_user(user)
    run "cd #{RABBITMQ_HOME_DIR} && #{sudo} runuser -l #{RABBITMQ_USER} -c '#{RABBITMQ_HOME_DIR}/sbin/rabbitmqctl set_permissions -p / #{user} \"^orchestrate.*\" \"^orchestrate.*|amq.default\" \"^orchestrate.*|amq.default\"'"
  end

  # reset the guest password and save it to 
  #
  # @param [String] password - the new password for guest
  def reset_guest_password(password)
    user = RABBITMQ_ADMIN
    change_password(user, password)
    write_user_password(user, password)
  end

  #write out a user and password to a file for use in adapters
  #TODO: CM - extract to base.rb and use this method from mongodb.rb when writing its passwords
  def write_user_password(user, password)
    password_file = RABBITMQ_PASSWORDS_FILE
    run "#{sudo} sh -c 'echo -e \"#{user}: #{password}\" >> #{password_file}'"
    run "#{sudo} chmod 600 #{password_file}"
  end

  # change the password for a user
  #
  # @param [String] user - the user to change the password for
  # @param [String] password - the new password
  def change_password(user, password)
    run "cd #{RABBITMQ_HOME_DIR} && #{sudo} runuser -l #{RABBITMQ_USER} -c '#{RABBITMQ_HOME_DIR}/sbin/rabbitmqctl change_password #{user} #{password}'"
  end

  # check if a user is already existing for rabbitmq
  #
  # @param [String] user - the user to lookup
  def user_exists?(user)
    output = @config.capture "cd #{RABBITMQ_HOME_DIR} && #{sudo} runuser -l #{RABBITMQ_USER} -c '#{RABBITMQ_HOME_DIR}/sbin/rabbitmqctl list_users'"
    lines = output.split("\r\n")[1..-1]

    !lines.empty? && lines.detect{|line| line.match(/\A#{user}\t.+/)}
  end
  
  # perform the backup data for rabbitmq (usually for a clean reinstall)
  # Folder will be located in the same location than the data with a timestamp
  #
  def backup_data
    installer.backup_folder_data("#{RABBITMQ_HOME_DIR}/var/lib/rabbitmq", RABBITMQ_USER)
  end


  # check the connection for rabbitmq from localhost(orchestrate) and from the remote point of view too
  #
  # Will raise an error if authentication fails or if tcp connection is wrong/not available
  # @param [String] host - the host to test rabbit is running on
  # @param [String] guest_password - the guest user's password to authenticate with
  def verify_connection(host, guest_password)
    # test to remote
    rabbit_available?(host: host, port: '5672', vhost: '/', username: RABBITMQ_ADMIN, password: guest_password)
  end

  # Tests to see if the RabbitMQ is accessible given the information the installer has been given
  # @param [Hash] rabbit_config The hash containig: host, port, username and password and vhost 
  # See ampq start/connect method for details on those keys
  def rabbit_available?(rabbit_config)

    t = Thread.new do
      host = rabbit_config[:host]
      port = rabbit_config[:port]
      opts = rabbit_config.merge(on_tcp_connection_failure: Proc.new { raise "Could not connect to #{host} #{port}"}, 
                                 on_possible_authentication_failure: Proc.new { raise "Auth Failed for #{rabbit_config[:username]} on #{host}"})
      AMQP.start(opts) do
        puts "Successful connection to #{host}!"
        EM.stop()
      end
    end

    t.join
  end

end

Capistrano.plugin :rabbit, Installer::Plugins::Rabbitmq