#
# Copyright 2013 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

module Installer::Plugins::Yum
  # Add Yum repository if not already setup
  #
  # @param [String] name the name of the repository to add
  # @param [Hash] options the set of options config for that repository. See default_repository_config_options for details
  def add_repository(name, options = {})
    name =  friendly_filename(name)
    name << '.repo' unless name.match /\.repo\Z/

    yum_filename = File.join(default_yum_dir_config_path, name)
    temporary_filename = File.join(user_home_path, name)
    if installer.remote_file_exists?(yum_filename)
      logger.info 'Repository already added. Skipping...'
    else
      options = default_repository_config_options.merge(options)
      content = repository_config_content(options)

      put content, temporary_filename
      run "#{sudo} mv #{temporary_filename} #{yum_filename}"
    end
  end

  # generate the content for a new repository configuration for Yum
  #
  # @param [Hash] options the different options for the config of a repository for yum
  #               see default_repository_config_options for details
  #
  # @return [String] the content of the config file generated
  def repository_config_content(options)
    <<-EOF.strip_heredoc 
      [#{options[:short_name]}]
      name=#{options[:name]}
      enabled=#{options[:enabled]}
      baseUrl=#{options[:base_url]}
      gpgcheck=#{options[:gpgcheck]}
    EOF
  end

  # Get the use home path
  #
  # @return [String] the home path for the current running user
  def user_home_path
    "/home/#{user}"
  end

  # Get the yum configuration directory path
  #
  # @return [String] defaault path to the yum repository config directory
  def default_yum_dir_config_path
    '/etc/yum.repos.d'
  end

  # Default repository options creation
  #
  # @return [Hash] default options for a repository config file
  # short_name: a name for the section config
  # name: name of the repository
  # enabled: 1 or 0. 1 if enabled
  # gpgcheck: 1 or 0. 1 if wanting to check gpg
  # base_url: Collabnet repository base url for the rpm packages
  def default_repository_config_options
    { short_name: COLLABNET_REPOSITORY_NAME,
      name: 'No Name',
      enabled: 1, 
      gpgcheck: 0, 
      base_url: fetch(:rpm_repo, nil) || 'http://orchestrate.collab.net/RHEL/$releasever/x86_64' }
  end

  # Return a friendly Unix filename. 
  # 1) It trims extra white space, 
  # 2) remove non alpha numeric beside _ and -
  # 3) remove repeated \t,\n and white space between words
  # 4) replace \t,\n and white space to _
  #
  # @param [String] filename  a filename that might not have clean naming
  # @return [String] a clean filename for a Posix system
  def friendly_filename(filename)
    filename.strip
            .gsub(/[^\w\s_-]+/, '')
            .gsub(/(^|\b\s)\s+($|\s?\b)/, '\\1\\2')
            .gsub(/\s+/, '_')
  end
  
  # Update packages specified or all installed packages if none specified.
  #
  # @param [Array] packages  List of packages to update. Default: all
  def update(packages = [])
    packages = [ packages ] unless packages.is_a?(Array)
    run "#{sudo} yum -y update #{packages.join(" ")}"
  end
    
  # Remove packages specified 
  #
  # @param [Array] packages  List of packages to remove. 
  def remove(packages)
    packages = [ packages ] unless packages.is_a?(Array)
    run "#{sudo} yum -y remove #{packages.join(" ")}"  
  end
  
  # Install via yum.
  # 
  # @param [Array] packages Packages to install, either String (for single package) or Array
  # @param [Boolean] update_existing If package exists, where to yum update it, defaults to true  
  def install(packages, update_existing = true)    
    packages = [ packages ] unless packages.is_a?(Array) 
    
    if update_existing
      installed_packages = packages_installed(packages)
      run "#{sudo} yum -y update #{installed_packages.join(" ")}" unless installed_packages.blank?
    end
    
    run "#{sudo} yum -y install #{packages.join(" ")}" unless packages.blank?
  end

  # Return the list of packages installed for the list specified
  def packages_installed(packages)
    installed_packages = []
      
    data = @config.capture "#{sudo} yum -d 0 list installed #{packages.join(" ")}"
    lines = data.split("\n")[1..-1]
    if lines.blank?
      logger.error data
    else
      installed_packages += lines.collect { |line| line.split(".").first }
    end
    packages & installed_packages 
  rescue
    []
  end
  
  # Clean yum
  def clean
    run "#{sudo} yum -y clean all"
  end
end

Capistrano.plugin :yum, Installer::Plugins::Yum