#
# Copyright 2015 CollabNet, Inc. All rights reserved. Usage governed by the LICENSE.txt file.
#

require 'active_support'
require 'active_support/core_ext'

class SiteOptions
  attr_reader :options
  attr_reader :filename

  KEY_SEPARATOR = '.'
  KEY_VALUE_SEPARATOR = /=/

  # @param [Array] files List of files to load; values in RHS taking precedence over those found in LHS files
  def initialize(files=[])
    @options = {}
    files.each { |filename|
      parse(open(filename).read) if filename.present? && File.exists?(filename)
    }
  end

  # Parse a multi-line string and create an options hash out of it.
  #
  # @param [String] body The contents to parse individual lines from and create an options hash
  # @return [Hash] The options hash created from the key[s]=value lines of the provided buffer
  def parse(body)
    result = {}
    self.class.parse(body) do |line_hash|
      result.deep_merge! line_hash
    end
    @options.deep_merge! result
  end

  # Parse a multi-line string and yield each individual hash constructed to the caller
  # @param [String] body The contents to parse individual lines from
  def self.parse(body)
    lines = body.split("\n").map(&:strip).reject { |line| line[0] == '#' || line.blank? }
    lines.each {|line| yield parse_line(line) }
  end

  # Convert a line containing a key[s]=value combination into a hash.  Keys and values are stripped.
  # @param [String] line The line containing the key[s]=value to convert.
  def self.parse_line(line)
    key, value = line.split(KEY_VALUE_SEPARATOR, 2).map(&:strip)
    if key.include? KEY_SEPARATOR
      parse_keyset(key, value)
    else
      value.blank? ? {} : { key.to_sym => value }
    end
  end

  # Convert a key set (key1.key2.key3) and value into a deep hash where each prior key points to a hash with the next
  # key, until the last whose key points to the value provided.
  #
  # @param [String] key A string containing the key set to construct a heirarchical hash from.
  # @param [String] value The end value to be placed at the bottom of the key hash.
  # @return [Hash] A deep hash where each key points to another hash, until the last key which points to the value.
  def self.parse_keyset(key, value)
    keys = key.split(KEY_SEPARATOR).map &:to_sym

    base_hash = intermediate = {}
    keys[0..-2].each do |sub_key|
      intermediate[sub_key] = {}
      intermediate = intermediate[sub_key]
    end
    intermediate[keys.last] = value unless value.blank?

    return base_hash
  end

  # Convert a deep hash into a site-options formatted string, suitable for writing to file.
  #
  # @param [Hash] data
  # @param [Boolean] sanitize Remove passwords from results
  # @param [Array] pfx Array to hold the key sequence to arrive at the current position in the hash.
  def self.to_options_string(data, sanitize=true, pfx=[])
    outdata = ""
    data.sort.map { |k, v|
      pfx << k
      if v.is_a? Hash
        outdata += to_options_string(v, sanitize, pfx)
      else
        outdata += "#{pfx.join('.')}=#{v}\n" unless sanitize && (k.end_with?('password') || k.end_with?('pass'))
      end
      pfx.pop
    }
    outdata
  end
end
