module CtfApiAuthentication

  # Simulate the OmniAuth Hash structure
  UserObject = Struct.new(:uid, :provider, :info)
  UserInfo = Struct.new(:name, :email)

  #
  def authenticate_with_teamforge_and_get_permissions
    orchestrate_session_id = request.headers["X-Orchestrate-Session"]
    @ctf_session = CtfService.create_teamforge_session(orchestrate_session_id)
    teamforge_service = @ctf_session.teamforge
    @ctf_user_data = teamforge_service.get_current_user_data
    get_user_permissions
  rescue TeamForge::ServiceError => ex
    logger.warn(ex.message)
    head status: 404
  end


  def set_reporting_access
    @privileged_projects = []
    ctf_projects = TeamForge::ProjectConfiguration
    # Associations viewer in CTF needs to be functioning on either Pipeline Read permission or Reporting API permission,
    # and not just the latter one. Below way of doing it avoids cascading changes to other methods called in the flow.
    @current_user.project_ids_with_permission(:pipeline_read).each { |orc_proj_id| @privileged_projects << ctf_projects.where(:project_id => orc_proj_id).first.ctf_project_id }
    @current_user.project_ids_with_permission(:reporting_api).each { |orc_proj_id| @privileged_projects << ctf_projects.where(:project_id => orc_proj_id).first.ctf_project_id }
    @privileged_projects.flatten.uniq
  end

  # @param [String] ctf_project_id
  # @return [Array] ARRAY of ctf project admins
  def get_project_admins(ctf_prj_id)
    ctf_ses = CtfService.create_authenticated_teamforge_session
    ctf_ser = ctf_ses.teamforge
    ctf_ser.list_project_admins(project_id: ctf_prj_id)[:multi_ref]
  end

  # Checks if the current user is an admin on a given project
  # @param [String] ctf_prj_id
  # @return [Boolean] true or false
  def is_project_admin?(ctf_prj_id)
    project_admins = get_project_admins(ctf_prj_id)
    if project_admins
      project_admins.any? { |admin| admin[:user_name] == current_user[:uid] }
    end
  end

  def get_projects_where_usr_is_admin
    admin_on_projects = []
    Project.each do |orc_proj|
      ctf_id = orc_proj.project_configuration.ctf_project_id
      admin_on_projects << ctf_id if is_project_admin?(ctf_id)
    end
    admin_on_projects
  end

  def get_user_permissions
    @ctf_user = @ctf_user_data[:multi_ref] || @ctf_user_data
    auth = UserObject.new.tap do |u|
      u.uid = @ctf_user[:username]
      u.provider = 'teamforge'
      u.info = UserInfo.new(:name => @ctf_user[:full_name], :email => @ctf_user[:email])
    end
    @current_user = User.find_or_create_user(auth)
    ctf_permissions_check
  end

  def check_post_schema_permissions_for_user
    if !(@current_user.super_user || @current_user.project_ids_with_permission(:pipeline_create).count > 0)
      raise Exception.new("POST_SCHEMA_PERMISSIONS_EXCEPTION. #{@current_user.inspect}")
    end
  rescue Exception => ex
    logger.error(ex.message)
    head status: 404
  end

  # Source API differs in its required permissions, all actions require:
  # :site_admin || :pipeline_edit (for current proj) || :pipeline_create (for current_proj) || :project_admin (for current_proj)
  def require_crud_permissions
    require_api_permission('SOURCE')
  end

  def current_user
    @current_user
  end

  def privileged_projects
    @privileged_projects
  end

  # common permission permutation
  # @return [Boolean] true or false
  # :site_admin || :pipeline_edit (for current proj) || :pipeline_create (for current_proj) || :project_admin (for current_proj)
  def can_access_resources?
    (current_user.super_user || current_user.has_write_permission_for_project_id?(current_project.id) || is_project_admin?(current_project.project_configuration.ctf_project_id))
  end

  # Tool API differs in its required permissions, all actions require:
  # :site_admin || :pipeline_edit (for current proj) || :pipeline_create (for current_proj) || :project_admin (for current_proj)
  def can_crud_tools?
    require_api_permission('TOOL')
  end

  # Tracker API differs in its required permissions, all actions require:
  # :site_admin || :pipeline_edit (for current proj) || :pipeline_create (for current_proj) || :project_admin (for current_proj)
  def require_crud_tracker
    require_api_permission('TRACKER')
  end

  # common api user permission
  def require_api_permission(resource='API')
    unless can_access_resources?
      raise Exception.new("#{resource}_PERMISSIONS_EXCEPTION. #{current_user.inspect}")
    end
  rescue Exception => ex
    logger.error(ex.message)
    head status: 404
  end

end
